/*
 * Copyright (c) 2003 by Hewlett-Packard Development Company, L.P.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE. 
 */ 


#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#define AO_REQUIRE_CAS
#include "atomic_ops.h"
#include "handler_safe_data.h"

/* A simple map data structure that may be accessed from a signal	*/
/* handler.  The map is represented as a chained hash table.  Nothing	*/
/* is ever removed from the table.  This makes it safe to insert at	*/
/* the front of a chain with compare-and-swap.				*/
/* The only update operations are to insert a new entry or to increment	*/
/* the value associated with an old one.				*/

struct HSD_map_entry {
  struct HSD_map_entry * volatile next;	/* Hash chain link */
  HSD_map_key_t key;
  HSD_map_data_t value;
};

#define HASH(x) (((unsigned long)x >> 13) ^ (unsigned long)x)

void
HSD_map_apply_all(HSD_map *p,
		   void (*f)(HSD_map_key_t, HSD_map_data_t, void *), void *data)
{
  int i;
  volatile struct HSD_map_entry *curr;

  for (i = 0; i < HSD_MAP_SIZE; ++i)
    for (curr = p -> hash[i]; curr != 0; curr = curr -> next)
      {
	f(curr -> key, curr -> value, data);
      }
}

/* 
 * Return a pointer to the unique HSD_map_entry for key.
 * If it doesn't exist, atomically insert it with a value of 0.
 */
static struct HSD_map_entry * get_entry(HSD_map *p, HSD_map_key_t key)
{
  unsigned long h = HASH(key);
  struct HSD_map_entry * volatile *prev = p -> hash + h;
  struct HSD_map_entry *curr = *prev;

  while (0 != curr)
    {
      if (curr -> key == key)
	return curr;
      prev = &(curr -> next);
      curr = *prev;
    } 
  {
    /* We know that key is not represented among the list nodes up to	*/
    /* the one containing the prev link.  The only way it can be 	*/
    /* added by another thread is if *prev becomes non-null while	*/
    /* we are doing this.						*/
    struct HSD_map_entry *nme = HSD_malloc(sizeof(struct HSD_map_entry));

    nme -> key = key;
    nme -> next = 0;	/* unnecessary? */
    nme -> value = 0;	/* unnecessary? */
    while (!AO_compare_and_swap_release((volatile AO_T *)prev,
			    		0, (AO_T)nme))
      {
        curr = *prev;
        while (0 != curr)
          {
            if (curr -> key == key)
	      {
		/* This may leak nme under unlikely conditions. */
		/* In the worst, astronomically unlikely, case	*/
		/* the leak is bound by the size of the data	*/
		/* structure times number of threads, since we	*/
		/* cant remove anything from the data structure.*/
      	        return curr;
	      }
            prev = &(curr -> next);
            curr = *prev;
          }
      }
    /* compare_and_swap succeeded */
    assert(*prev == nme && nme -> key == key);
    return nme;
  }
}

/* 
 * Return the value of map p at key.  Map values that were not
 * explicitly set or incremented are zero.
 */
HSD_map_data_t
HSD_map_get(HSD_map *p, HSD_map_key_t key)
{
  struct HSD_map_entry *q = get_entry(p, key);

  return (HSD_map_data_t) AO_load((AO_T *)(&(q -> value)));
}


/* 
 * Increment the value of map p at key by n.  If it was not previously set
 * its value becomes n.  Return the old value.
 */
HSD_map_data_t
HSD_map_add(HSD_map *p, HSD_map_key_t key, long n)
{
  struct HSD_map_entry *q = get_entry(p, key);

  return AO_fetch_and_add((AO_T *)(&(q -> value)), n);
}

/* 
 * Set the value of map p at key to v.
 */
void
HSD_map_set(HSD_map *p, HSD_map_key_t key, HSD_map_data_t n)
{
  struct HSD_map_entry *q = get_entry(p, key);

  AO_store((AO_T *)(&(q -> value)), n);
}
